import os
import json
import uuid
from typing import Union

from .NBT import NBT

path = os.path.dirname(os.path.abspath(__file__))
with open(f'{path}/../metadata/data/entity_nbt.json', 'r') as file:
    entities = json.load(file)

entities_nbt = {}
for key in entities:
    entities_nbt[key] = list(entities[key].keys())

def generate_signed_32bit_integer_uuid_list():
    # Generate a version 4 UUID
    uuid_v4 = uuid.uuid4()

    # Convert the UUID to a 128-bit integer
    uuid_int = uuid_v4.int

    # Create a list to store the 4 32-bit integers
    uuid_32bit_list = []

    # Mask and shift the 128-bit integer to obtain 4 32-bit integers
    for i in range(4):
        mask = 0xFFFFFFFF << (32 * i)
        uuid_32bit = (uuid_int & mask) >> (32 * i)

        # Convert the 32-bit unsigned integer to a signed integer
        if uuid_32bit >= 2**31:
            uuid_32bit -= 2**32

        uuid_32bit_list.append(uuid_32bit)

    return uuid_32bit_list[::-1]  # Reverse the list to get the correct order
   
class Entity:
    def __init__(self, name: str, nbt: Union[NBT, dict, None] = None):
        
        name = name.lower()
        
        if name not in entities.keys():
            raise ValueError(f"Entity '{name}' does not exist")
        
        self.__dict__['descriptor'] = name
        
        if type(nbt) == dict:
            self.__dict__['nbt'] = NBT(nbt)
        elif type(nbt) == NBT:
            self.__dict__['nbt'] = nbt
        elif nbt == None:
            self.__dict__['nbt'] = NBT()
        else:
            raise ValueError(f"nbt must be of type NBT or dict, not {type(nbt)}")

        self.__dict__['nbt'].UUID = generate_signed_32bit_integer_uuid_list()
        
    @property
    def attributes(self):
        return entities_nbt[self.descriptor]
    
    def update(self, nbt):
        from shulker.functions.update_entity import update
        
        uuid = self.__getattr__("UUID")
        return update(uuid, nbt)
    
    def summon(self, *args):
        from shulker.functions.default import summon
        
        if len(args) == 3:
            coords = (args[0], args[1], args[2])
        elif len(args) == 1:
            coords = args[0]
        else:
            raise ValueError(f"summon takes 1 (tuple, list, Coordinates...) or 3 (x, y, z) arguments, not {len(args)}")
        
        return summon(self, coords)
    
    def __setattr__(self, name, value):
        if name in self.attributes:
            self.nbt.__setattr__(name, value)
        else:
            raise ValueError(f"Attribute '{name}' for '{self.descriptor}' does not exist")

    def __getattr__(self, name):
        return getattr(self.nbt, name)
        
    def __str__(self):
        if hasattr(self, "nbt") and self.nbt not in [None, "", "{}"]:
            return f"{self.descriptor}{self.nbt}"
        else:
            return f"{self.descriptor}"




from typing import Union

from .Coordinates import Coordinates
from .BlockCoordinates import BlockCoordinates

class Zone:
    """
    Custom component that is a set of two Coordinates(), representing an area
    """

    def __init__(
        self, pos1: Union[Coordinates, tuple], pos2: Union[Coordinates, tuple]
    ):

        if isinstance(pos1, tuple):
            self.pos1 = Coordinates(*pos1)
        else:
            self.pos1 = pos1

        if isinstance(pos2, tuple):
            self.pos2 = Coordinates(*pos2)
        else:
            self.pos2 = pos2

    def contains(self, coord: Coordinates) -> bool:
        """
        Checks if the given Coordinates is within the Zone.
        """
        if not isinstance(coord, Coordinates):
            raise ZoneWrongCoordsType(
                "Zone contains method requires a Coordinates instance as input"
            )

        min_x = min(self.pos1.x, self.pos2.x)
        max_x = max(self.pos1.x, self.pos2.x)
        min_y = min(self.pos1.y, self.pos2.y)
        max_y = max(self.pos1.y, self.pos2.y)
        min_z = min(self.pos1.z, self.pos2.z)
        max_z = max(self.pos1.z, self.pos2.z)

        return (min_x <= coord.x <= max_x) and (min_y <= coord.y <= max_y) and (min_z <= coord.z <= max_z)
    
    def offset(self, x=0, y=0, z=0):
        """
        Offsets the coordinates by the given tuple and returns a new Zone object.
        """
        return Zone(self.pos1.offset(x, y, z), self.pos2.offset(x, y, z))
        
    def __str__(self):
        if not isinstance(self.pos1, Coordinates) or not isinstance(
            self.pos2, Coordinates
        ):
            raise ZoneWrongCoordsType(
                "A Zone requires to be provided a set of two Coordinates instances as pos1 and pos2"
            )

        return f"{self.pos1} {self.pos2}"

class ZoneWrongCoordsType(Exception):
    pass



from typing import Union


class Coordinates:
    def __init__(
        self,
        x: Union[int, float, str],
        y: Union[int, float, str],
        z: Union[int, float, str],
        yaw: Union[int, float, str] = None,
        pitch: Union[int, float, str] = None,
    ):

        self.x = x
        self.y = y
        self.z = z
        self.yaw = yaw
        self.pitch = pitch

    def check_carets(self):
        coords = [self.x, self.y, self.z]

        if any(str(coord).startswith("^") for coord in coords):
            if not all(str(coord).startswith("^") for coord in coords):
                raise WrongCaretNotation(
                    "When using caret notation, every coordinates must start with a caret"
                )

    def offset(self, x=0, y=0, z=0, yaw=0, pitch=0) -> "Coordinates":
        """
        Offsets the coordinates by the given tuple and returns a new BlockCoordinates object.
        """
        
        if yaw and pitch and self.yaw and self.pitch:
          return Coordinates(self.x + x, self.y + y, self.z + z, self.yaw + yaw, self.pitch + pitch)
        if pitch and self.pitch:
          return Coordinates(self.x + x, self.y + y, self.z + z, self.pitch + pitch)
        if yaw and self.yaw:
          return Coordinates(self.x + x, self.y + y, self.z + z, self.yaw + yaw)
        
        return Coordinates(self.x + x, self.y + y, self.z + z)
        
    def __str__(self):
        self.check_carets()
        
        if self.yaw is not None and self.pitch is not None:
          return f"{self.x} {self.y} {self.z} {self.yaw} {self.pitch}"
        
        return f"{self.x} {self.y} {self.z}"


class WrongCaretNotation(Exception):
    pass



import unittest

from shulker.components.TargetSelector import TargetSelector
from shulker.components.TargetSelector import (
    IncorrectTargetSelectorIdentifier,
    IncorrectTargetSelectorArgumentsType,
    InvalidTargetSelectorArgumentKey,
)


class TestBlock(unittest.TestCase):
    def test_valid_identifier(self):
        """Test if valid identifiers are correctly processed."""
        ts = TargetSelector('p')
        self.assertEqual(str(ts), '@p[]')
        ts = TargetSelector('a')
        self.assertEqual(str(ts), '@a[]')
        ts = TargetSelector('r')
        self.assertEqual(str(ts), '@r[]')
        ts = TargetSelector('s')
        self.assertEqual(str(ts), '@s[]')
        ts = TargetSelector('e')
        self.assertEqual(str(ts), '@e[]')

    def test_invalid_identifier(self):
        """Test if invalid identifiers raise the proper exception."""
        with self.assertRaises(IncorrectTargetSelectorIdentifier):
            TargetSelector('invalid')

    def test_valid_arguments(self):
        """Test if valid arguments are correctly processed."""
        ts = TargetSelector('e', {'type': 'cow', 'limit': 1})
        self.assertEqual(str(ts), '@e[type=cow,limit=1]')

    def test_invalid_argument_type(self):
        """Test if invalid argument types raise the proper exception."""
        with self.assertRaises(IncorrectTargetSelectorArgumentsType):
            TargetSelector('e', "type=cow,limit=1")

    def test_invalid_argument_key(self):
        """Test if invalid argument keys raise the proper exception."""
        with self.assertRaises(InvalidTargetSelectorArgumentKey):
            ts = TargetSelector('e', {'invalid_key': 'value'})
            str(ts)

    def test_list_arguments(self):
        """Test if list arguments are correctly processed."""
        ts = TargetSelector('a', {'tag': ['has_tag', '!hasnt_tag']})
        self.assertEqual(str(ts), '@a[tag=has_tag,tag=!hasnt_tag]')

    def test_empty_arguments(self):
        """Test if empty arguments are correctly processed."""
        ts = TargetSelector('s')
        self.assertEqual(str(ts), '@s[]')

    def test_no_arguments(self):
        """Test if None arguments are correctly processed."""
        ts = TargetSelector('r', None)
        self.assertEqual(str(ts), '@r[]')


if __name__ == "__main__":
    unittest.main()



import unittest

from shulker.components.BlockState import (
    BlockState,
    UnexpectedBlockStatePropertyValueType,
)
from shulker.components.Block import Block


class TestBlock(unittest.TestCase):
    def test_blockstate_without_args(self):
        """Test creating an empty BlockState with no arguments."""
        bs = BlockState()

        self.assertEqual(str(bs), "[]")

    def test_blockstate_without_attrs(self):
        """Test BlockState string representation when no attributes are set."""
        b = Block("dirt")
        
        self.assertEqual(str(b.blockstate), "[]")

    def test_blockstate_without_args_passed_to_Block(self):
        """Test BlockState passed as an argument to the Block class without arguments."""
        bs = BlockState()
        b = Block("dirt", blockstate=bs)

        self.assertEqual(str(b.blockstate), "[]")

    def test_blockstate_without_args_attributed_to_Block(self):
        """Test BlockState attributed to the Block class without arguments."""
        bs = BlockState()
        b = Block("dirt")
        b.blockstate = bs

        self.assertEqual(str(b.blockstate), "[]")

    def test_blockstate_with_arg(self):
        """Test creating a BlockState with a single property."""
        bs = BlockState({"snowy": True})

        diff = str(bs)
        test = "[snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_with_args(self):
        """Test creating a BlockState with multiple properties."""
        bs = BlockState({"snowy": True, "facing": "north"})

        diff = str(bs)
        test = "[facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_via_attr(self):
        """Test setting a single property via attribute assignment."""
        bs = BlockState()
        bs.snowy = True

        diff = str(bs)
        test = "[snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_via_attrs(self):
        """Test setting multiple properties via attribute assignments."""
        bs = BlockState()
        bs.snowy = True
        bs.facing = "north"

        diff = str(bs)
        test = "[facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_with_arg_passed_to_Block(self):
        """Test passing a BlockState with a single property to the Block class."""
        bs = BlockState({"snowy": True})
        b = Block("dirt", blockstate=bs)

        diff = str(b.blockstate)
        test = "[snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_with_args_passed_to_Block(self):
        """Test passing a BlockState with multiple properties to the Block class."""
        bs = BlockState({"snowy": True, "facing": "north"})
        b = Block("dirt", blockstate=bs)

        diff = str(b.blockstate)
        test = "[facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_via_attr_through_Block(self):
        """Test setting a single property through the Block class."""
        b = Block("dirt")
        b.blockstate.snowy = True

        diff = str(b.blockstate)
        test = "[snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_via_attrs_through_Block(self):
        """Test setting multiple properties through the Block class."""
        b = Block("dirt")
        b.blockstate.snowy = True
        b.blockstate.facing = "north"

        diff = str(b.blockstate)
        test = "[facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_mixed_attrs_and_args_through_Block(self):
        """Test a combination of properties set via attributes and arguments through the Block class."""
        bs = BlockState({"snowy": True, "dancing": "well"})
        b = Block("dirt")
        b.blockstate = bs
        b.blockstate.facing = "north"

        diff = str(b.blockstate)
        test = "[dancing=well,facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_override_attr_through_Block(self):
        """Test overriding a property through the Block class."""
        bs = BlockState({"snowy": True})
        b = Block("dirt")
        b.blockstate = bs
        b.blockstate.snowy = False

        diff = str(b.blockstate)
        test = "[snowy=false]"
        self.assertEqual(diff, test)

    def test_blockstate_incorrect_argument_value_type(self):
        """Test BlockState with an incorrect value type for a property."""
        bs = BlockState({"snowy": ["test", "test1"]})

        with self.assertRaises(UnexpectedBlockStatePropertyValueType):
            str(bs)
            
    def test_blockstate_non_string_key(self):
        """Test BlockState with a non-string key in the input dictionary."""
        with self.assertRaises(ValueError):
            bs = BlockState({123: True})



import os
import json

from .BlockState import BlockState
from .BlockCoordinates import BlockCoordinates
from .Coordinates import Coordinates
from .NBT import NBT

path = os.path.dirname(os.path.abspath(__file__))
with open(f'{path}/../metadata/data/block_list.json', 'r') as file:
    blocks_data = json.load(file)

blocks = {}
for key in blocks_data:
    if 'properties' in blocks_data[key]:
        blocks[key] = {}
        blocks[key]['properties'] = blocks_data[key]['properties']
        for states in blocks_data[key]['states']:
            if 'default' in states.keys():
                blocks[key]['default'] = states['properties']
    else:
        blocks[key] = {'properties': {}, 'default': {}}

class Block:
    """
    A single '<block>' argument looks like this:

        'stone'
        'minecraft:redstone_wire[power=15,north=up,south=side]'
        'minecraft:jukebox{RecordItem:{...}}'
        'minecraft:furnace[facing=north]{BurnTime:200}'

    The format of '<block>' parameters is 'namespaced_ID[block_states]{data_tags}', in which
    block states and data tags can be omitted when they are not needed.

        Namespaced ID is required (though if namespace isn't set it defaults to 'minecraft:').
            In the context of "conditions"/testing for blocks, it can also be the namespace ID
            of block tag with the prefix of '#', such as '#minecraft:planks'.

        Block states are inside '[]', comma-separated and must be properties/values supported
        by the blocks. They are optional.
            'minecraft:stone[doesntexist=purpleberry]' is a syntax error, because 'stone'
            doesn't have 'doesntexist'.
            'minecraft:redstone_wire[power=tuesday]' is a syntax error, because 'redstone_wire''s
            'power' is a number between 0 and 15.

        Data tags are inside '{}'. It's optional.

        In the context of "conditions"/testing for blocks, only the states provided are tested.
            If command tests 'redstone_wire[power=15]', it checks only power, but ignores other
            states such as north.

        In the context of setting blocks, any states provided are set, but anything omitted retain
        their default values, depending on the block.
            If command sets 'redstone_wire[power=15]', it is set 'power' to 15, but 'north' is a
            default value (in this case, set to 'none').
    """

    def __init__(self, id: str, blockstate: BlockState = None, nbt: NBT = None):

        if not isinstance(id, str) or id == "":
            raise ValueError("id must be a non empty string")

        if ":" in id:
            parsed = id.split(":")
            self.__dict__['namespace'] = parsed[0]
            self.__dict__['id'] = parsed[1]
        else:
            self.__dict__['namespace'] = "minecraft"
            self.__dict__['id'] = id.replace(" ", "_")
            
        self.__dict__['descriptor'] = self.__dict__['namespace'] + ":" + self.__dict__['id']
            
        if self.__dict__['namespace'] == "minecraft":
            if self.__dict__['descriptor'] not in blocks:
                raise ValueError(f"Unknown block ID: {id}")

        self.__dict__['coords'] = None
        
        if self.__dict__['namespace'] == "minecraft":
            self.__dict__['blockstate'] = (
                blockstate if isinstance(blockstate, BlockState) else BlockState(blocks[self.descriptor]['default'])
            )
        else:
            self.__dict__['blockstate'] = (
                blockstate if isinstance(blockstate, BlockState) else BlockState({})
            )
        self.__dict__['nbt'] = nbt if isinstance(nbt, NBT) else NBT(nbt)

    @property
    def attributes(self):
        return blocks[self.descriptor]['properties']
    
    def set(self, *args, handler=None):
        from shulker.functions.default import set_block
        if len(args) == 3:
            coords = (args[0], args[1], args[2])
        elif len(args) == 1:
            coords = args[0]
        else:
            raise ValueError(f"set takes 1 (tuple, list, Coordinates...) or 3 (x, y, z) arguments and eventually an handler=.., not {len(args)}")
        
        self.__dict__['coords'] = coords
        
        if handler:
            return set_block(coords, self, handler)
        else:
            return set_block(coords, self)
    
    def move(self, *args, handler=None, replace_by="air"):
        from shulker.functions.default import set_block
        if len(args) == 3:
            new_coords = (args[0], args[1], args[2])
        elif len(args) == 1:
            new_coords = args[0]
        else:
            raise ValueError(f"move takes 1 (tuple, list, Coordinates...) or 3 (x, y, z) arguments and eventually an handler=.. and replace_by=.., not {len(args)}")
        
        old_coords = self.__dict__['coords']
        self.__dict__['coords'] = new_coords
        
        if handler:
            set_block(old_coords, replace_by, handler)
            set_block(new_coords, self, handler)
        else:
            set_block(old_coords, replace_by)
            set_block(new_coords, self)
    
    def offset(self, x=0, y=0, z=0, handler=None, replace_by="air"):

        if type(self.__dict__['coords']) in [tuple, list]:
            to_offset = BlockCoordinates(*self.__dict__['coords'])
        elif type(self.__dict__['coords']) in [BlockCoordinates, Coordinates]:
            to_offset = self.__dict__['coords']
            
        new_coords = to_offset.offset(x=x, y=y, z=z)
        
        self.move(new_coords, handler=handler, replace_by=replace_by)
        
    def __setattr__(self, name, value):
        if name == "nbt":
            if type(value) == NBT:
                self.__dict__['nbt'] = value
            else:
                self.__dict__['nbt'] = NBT(value)
        elif name == "blockstate":
            if type(value) == BlockState:
                self.__dict__['blockstate'] = value
            else:
                self.__dict__['blockstate'] = BlockState(value)
        elif name in self.attributes and value in self.attributes[name]:
            setattr(self.blockstate, name, value)
        else:
            raise ValueError(f"BlockState '{name}={value}' for '{self.descriptor}' does not exist")
    
    def __getattr__(self, name):
        if name in self.attributes:
            return getattr(self.blockstate, name)
        else:
            return getattr(self, name)
    
    def __str__(self):
        return f"{self.descriptor}{self.blockstate}{self.nbt}"



import re
from typing import Union

from shulker.functions.base_functions import *
from shulker.components.Coordinates import Coordinates
from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.Block import Block
from shulker.components.BlockHandler import BlockHandler
from shulker.components.Entity import Entity
from shulker.components.BlockZone import BlockZone

############ FILL ############

def meta_set_zone(
    zone: BlockZone, block: Block, handler: BlockHandler, filter: Union[Block, None] = None
) -> str:

    if str(handler) == "replace" and filter:
        handler = str(handler) + " " + str(filter)

    return f"fill {zone} {block} {handler}"


def set_zone(
    zone: Union[BlockZone, list],
    block: Union[Block, str],
    handler: Union[BlockHandler, str] = "replace",
    filter: Union[Block, str] = "",
) -> str:
    """
    Available handlers:
        'replace' — The old block drops neither itself nor any contents. Plays no sound.
        'destroy' — The old block drops both itself and its contents (as if destroyed by a player). Plays the appropriate block breaking noise.
        'keep' — Only air blocks are changed (non-air blocks are unchanged).
        'hollow' - Replaces only the blocks on the outer edge of the fill region with the
        specified block. Inner blocks are changed to air, dropping their contents as entities
        but not themselves. If the fill region has no inner blocks (because it is smaller than
        three blocks in at least one dimension), acts like 'replace'.
        'outline' - Replaces only the blocks on the outer edge of the fill region with the
        specified block. Inner blocks are not affected. If the fill region has no inner blocks
        (because it is smaller than three blocks in at least one dimension), acts like 'replace'.

    Defaults to 'replace'
    """

    check_output_channel()

    zone = format_arg(zone, BlockZone)
    block = format_arg(block, Block)
    handler = format_arg(handler, BlockHandler)

    if filter != "":
        filter = format_arg(filter, Block)

    cmd = meta_set_zone(zone, block, handler, filter)

    status = post(cmd)

    return status

############ SETBLOCK ############

def meta_set_block(coords: BlockCoordinates, block: Block, handler: Union[BlockHandler, None]) -> str:
    return f"setblock {coords} {block}{(' ' + str(handler)) if handler else ''}"


def set_block(
    coords: Union[BlockCoordinates, tuple],
    block: Union[Block, str],
    handler: Union[BlockHandler, str] = "replace",
) -> str:
    """
    Available handlers:
        'replace' — The old block drops neither itself nor any contents. Plays no sound.
        'destroy' — The old block drops both itself and its contents (as if destroyed by a player). Plays the appropriate block breaking noise.
        'keep' — Only air blocks are changed (non-air blocks are unchanged).

    Defaults to 'replace'
    """

    check_output_channel()

    coords = format_arg(coords, BlockCoordinates)
    block = format_arg(block, Block)
    handler = format_arg(handler, BlockHandler)

    cmd = meta_set_block(coords, block, handler)

    status = post(cmd)

    return status

############ SUMMON ############

def meta_summon(entity: str, coords: BlockCoordinates) -> str:
    return f"summon {entity.descriptor} {coords} {entity.nbt}"


def summon(
    entity: Union[Entity, str],
    coords: Union[BlockCoordinates, Coordinates, tuple],
) -> str:
    """Summons an entity at coords, can be provided nbt_data"""

    check_output_channel()
      
    entity = format_arg(entity, Entity)
    coords = format_arg(coords, Coordinates)
        
    cmd = meta_summon(entity, coords)

    status = post(cmd)
    
    return status


############ SAY ############

def meta_say(text: str) -> str:
    return f"say {text}"

def say(text: str) -> str:
    """Sends a message in the chat"""
    check_output_channel()
    
    cmd = meta_say(text)
    
    status = post(cmd)
    
    return status

######## BAN & KICK ########

def meta_ban(target: str, reason: str) -> str:
    return f"ban {target} {reason}"

def meta_banlist(option: str) -> str:
    return f"banlist {option}"

def ban(target: str, reason: Union[str, None]) -> str:
    """This function bans a player"""
    check_output_channel()
    
    if reason is None:
        reason = ""
        
    cmd = meta_ban(target, reason)
    
    status = post(cmd)
    
    return status
 
def ban_ip(target: str, reason: str) -> str:
    """This function bans an IP"""
    check_output_channel()
    
    cmd = meta_ban(target, reason)
    
    status = post(cmd)
    
    return status
        
def banlist(option: str) -> str:
    """Fetches the banlist"""
    check_output_channel()
    
    cmd = meta_banlist(option)
    
    status = post(cmd)
    
    return status

def meta_kick(target: str, reason: str) -> str:
    return f"kick {target} {reason}"

def kick(target: str, reason: Union[str, None]) -> str:
    """Kicks a player"""
    check_output_channel()
    
    if reason is None:
        reason = ""
        
    cmd = meta_kick(target, reason)
    
    status = post(cmd)
    
    return status

def meta_pardon(target: str) -> str:
    return f"pardon {target}"

def meta_pardon_ip(target: str) -> str:
    return f"pardon-ip {target}"

def pardon(target: str) -> str:
    """Unbans a player"""
    check_output_channel()
    
    cmd = meta_pardon(target)
    
    status = post(cmd)
    
    return status

def pardon_ip(target: str) -> str:
    """This function unbans an IP"""
    check_output_channel()
    
    cmd = meta_pardon_ip(target)
    
    status = post(cmd)
    
    return status

############ OP ############

def meta_op(target: str) -> str:
    return f"op {target}"

def meta_deop(target: str) -> str:
    return f"deop {target}"

def op(target: str) -> str:
    """Gives an operator status to a player"""
    check_output_channel()
    
    cmd = meta_op(target)
    
    status = post(cmd)
    
    return status
        
def deop(target: str) -> str:
    """Removes an operator status from a player"""
    check_output_channel()
    
    cmd = meta_deop(target)
    
    status = post(cmd)
    
    return status
        
############# SEED #############

def meta_seed() -> str:
    return f"seed"

def seed() -> str:
    """Fetches the seed of the world and return it as a string"""
    check_output_channel()
    
    cmd = meta_seed()
    
    status = post(cmd)
    
    seed = re.match(r"Seed: \[(-{0,1}(\d+))\]", status)
    return seed.group(1)

############# DIFFICULTY #############

def meta_set_difficulty(option: str) -> str:
    return f"difficulty {option}"

def meta_get_difficulty() -> str:
    return f"difficulty"

def set_difficulty(option: str) -> str:
    """Sets the difficulty of the world"""
    check_output_channel()
    
    cmd = meta_set_difficulty(option)
    
    status = post(cmd)
    
    return status
        
def get_difficulty() -> Union[bool, str]:
    """Fetches the difficulty of the world"""
    check_output_channel()
    
    cmd = meta_get_difficulty()
    
    status = post(cmd)
    
    difficulty = re.match(r"The difficulty is (\w+)", status)
    return difficulty.group(1)

########### WEATHER ###########

def meta_weather(option: str) -> str:
    return f"weather {option}"

def weather(option: str) -> str:
    """Sets the weather of the world
    Options: clear, rain, thunder"""
    
    check_output_channel()
    
    cmd = meta_weather(option)
    
    status = post(cmd)
    
    return status
        
########### MSG ###########

def meta_msg(target: str, message: str) -> str:
    return f"msg {target} {message}"
  
def msg(target: str, message: str) -> str:
    """Sends a private message to a player"""
    
    check_output_channel()
    
    cmd = meta_msg(target, message)
    
    status = post(cmd)
    
    return status

########### GAMEMODE ###########

def meta_default_gamemode(option: str) -> str:
    return f"defaultgamemode {option}"

def meta_gamemode(target: str, option: str) -> str:
    return f"gamemode {option} {target}"

def gamemode(target: str, option: str) -> str:
    """Sets the gamemode of a player
    Available options: survival, creative, adventure, spectator"""
    
    check_output_channel()
    
    cmd = meta_gamemode(target, option)
    
    status = post(cmd)
    
    return status
        
def default_gamemode(option: str) -> Union[bool, str]:
    """Sets the default gamemode of the world
    Available options: survival, creative, adventure, spectator"""
    
    check_output_channel()
    
    cmd = meta_default_gamemode(option)
    
    status = post(cmd)
    
    return status
    
########### TIME ###########

def meta_query_time(option: str) -> str:
    return f"time query {option}"

def query_time(option: str = "day") -> str:
    """Fetches the time of the world and returns it as a string
    Available options: day, daytime, gametime"""
    
    check_output_channel()
    
    cmd = meta_query_time(option)
    
    status = post(cmd)
        
    time = re.match(r"The time is (\d+)", status)
    return time.group(1)

def meta_add_time(value: str, option: str) -> str:
    return f"time add {value}{option}"

def add_time(value: int, option: str = "tick") -> str:
    """Adds time to the world
    Available options: day, second, tick"""
    
    check_output_channel()
    
    option = option[0]
    
    cmd = meta_add_time(value, option)
    
    status = post(cmd)
    
    return status
        
def meta_set_time(value: int, option: str) -> str:
    return f"time set {value}{option}"

def set_time(value: int, option: str = "tick") -> str:
    """Sets the time of the world
    Available options: day, second, tick"""
    
    check_output_channel()
    
    option = option[0]
    
    cmd = meta_set_time(value, option)
    
    status = post(cmd)
    
    return status
        
def meta_time(option: str) -> str:
    return f"time set {option}"

def time(option: str) -> str:
    """Sets the time of the world
    Available options: day, midnight, night, noon"""
    
    check_output_channel()
    
    cmd = meta_time(option)
    
    status = post(cmd)
    
    return status
        
########### EXPERIENCE ###########

def meta_xp_query(target: str, option: str) -> str:
    return f"xp query {target} {option}"

def xp_query(target: str, option: str) -> str:
    """Fetches the experience of a player and return it as a string
    Available options: levels, points"""
    
    check_output_channel()
    
    cmd = meta_xp_query(target, option)
    
    status = post(cmd)
    
    xp = re.findall(r"(\d+) experience " + option, status)
    return xp[0]

########### WHITELIST ###########

def meta_get_whitelist() -> str:
    return "whitelist list"

def get_whitelist() -> list:
    """Fetches the whitelist of the world and return as a list"""
    
    check_output_channel()
    
    cmd = meta_get_whitelist()
    
    status = post(cmd)[:-4]
        
    whitelist = status.split("players: ")[1].split(", ")
    return whitelist

def meta_toggle_whitelist(option: str) -> str:
    return f"whitelist {option}"

def toggle_whitelist(option: str) -> str:
    """Toggles the whitelist of the world
    Available options: on, off"""
    
    check_output_channel()
    
    cmd = meta_toggle_whitelist(option)
    
    status = post(cmd)
    
    return status
        
def meta_reload_whitelist() -> str:
    return f"whitelist reload"

def reload_whitelist() -> str:
    """Reloads the whitelist of the world"""
    
    check_output_channel()
    
    cmd = meta_reload_whitelist()
    
    status = post(cmd)
    
    return status
        
def meta_update_whitelist(option: str, target: str) -> str:
    return f"whitelist {option} {target}"

def update_whitelist(option: str, target: str) -> str:
    """Updates the whitelist of the world
    Available options: add, remove"""
    
    check_output_channel()
    
    cmd = meta_update_whitelist(option, target)
    
    status = post(cmd)
    
    return status
        
############ ADMIN ############

def meta_stop() -> str:
    return "stop"

def stop() -> str:
    """Stops the server"""
    
    check_output_channel()
    
    cmd = meta_stop()
    
    status = post(cmd)
    
    return status
        
def meta_save_all() -> str:
    return "save-all"

def save_all() -> Union[bool, str]:
    """This function saves the world"""
    
    check_output_channel()
    
    cmd = meta_save_all()
    
    status = post(cmd)
    
    return status
        
def meta_toggle_save(option: str) -> str:
    return f"save-{option}"

def toggle_save(option: str) -> str:
    """Toggles the auto-save of the world
    Available options: on, off"""
    
    check_output_channel()
    
    cmd = meta_toggle_save(option)
    
    status = post(cmd)
    
    return status
        
############# HELP #############

def meta_help(value: str) -> str:
    return f"help {value}"

def parse_help(prepender = ""):
    
    cmd = meta_help(prepender)
    
    data = post(cmd)
        
    # Remove all \x1b[.. color code from the status
    data = re.sub(r"\x1b\[[0-9;]*m", "", data)
    
    # Find the number of help pages
    nb_pages = re.findall(r" \(\d+/(\d+)\)", data)
    
    if nb_pages:
        nb_pages = nb_pages[0]
    elif " Help: " in data and not nb_pages:
        nb_pages = 1
    else:
        raise Exception("No help found")
        
    help_cmds = {}
    for page in range(1, int(nb_pages) + 1):
                
        cmd = meta_help(prepender + str(page))
        
        data = post(cmd)
            
        data = re.sub(r"\x1b\[[0-9;]*m", "", data)
        lines = data.splitlines()[(1 if page > 1 else 2):]
        
        for line in lines:
            if not line.startswith("/") and ":" not in line:
                continue
            splitted = line.split(": ")
            command = splitted[0]
            value = splitted[1]
            help_cmds[command] = value
    
    return help_cmds

def help(value: str = "") -> dict:
    """Fetches the help of the world and return a dict of it
    If a value is specified it will fetch the help of the value only"""
    
    check_output_channel()

    cmds = parse_help(value)
    for key in cmds:
        if not key.startswith('/'):
            cmds[key] = parse_help(prepender=key + " ")
            
    return cmds

############# LIST #############

def meta_list_players(uuids: bool = False) -> str:
    if uuids:
        uuids = "uuids"
    else:
        uuids = ""
    return f"list {uuids}"

def list_players(uuids: bool = False) -> list:
    """Fetches the list of players in the world
    If uuids is set to True it will be a list of uuids returned"""
    
    check_output_channel()
    
    cmd = meta_list_players(uuids)
    
    status = post(cmd)
    
    # Remove all \x1b[.. color code from the status
    status = re.sub(r"\x1b\[[0-9;]*m", "", status)
    
    players = status.split("players online: ")[1].split(", ")
        
    return players

############# SPECTATE #############

def meta_spectate(target: str, player: str) -> str:
    return f"spectate {target} {player}"

def spectate(target: str, player: str) -> str:
    """Makes the target spectates a player"""
    
    check_output_channel()
    
    cmd = meta_spectate(target, player)
    
    status = post(cmd)
    
    return status
        
        
############# SET WORLD SPAWN #############

def meta_set_world_spawn(pos: Coordinates, yaw: int) -> str:
    return f"setworldspawn {pos} {yaw}"

def set_world_spawn(pos: Union[Coordinates, tuple], yaw: int = 0) -> str:
    """This function sets the world spawn"""
    
    check_output_channel()
    
    pos = format_arg(pos, Coordinates)

    cmd = meta_set_world_spawn(pos, yaw)
    
    status = post(cmd)
    
    return status


import unittest

from shulker.components.Coordinates import Coordinates, WrongCaretNotation

class TestCoordinates(unittest.TestCase):
    def test_init(self):
        """Checks if the Coordinates class initializes correctly."""
        
        coords = Coordinates(1, 2, 3)
        self.assertEqual(coords.x, 1)
        self.assertEqual(coords.y, 2)
        self.assertEqual(coords.z, 3)
        self.assertIsNone(coords.yaw)
        self.assertIsNone(coords.pitch)

    def test_init_with_yaw_and_pitch(self):
        """Checks if the Coordinates class initializes correctly."""
        
        coords = Coordinates(1, 2, 3, 90, 45)
        self.assertEqual(coords.x, 1)
        self.assertEqual(coords.y, 2)
        self.assertEqual(coords.z, 3)
        self.assertEqual(coords.yaw, 90)
        self.assertEqual(coords.pitch, 45)

    def test_check_carets_valid(self):
        """Verifies that check_carets() method works correctly."""
        
        coords = Coordinates("^1", "^2", "^3")
        try:
            coords.check_carets()
        except WrongCaretNotation:
            self.fail("check_carets() raised WrongCaretNotation unexpectedly")

    def test_check_carets_invalid(self):
        """Verifies that check_carets() method raises an exception."""
        
        coords = Coordinates(1, "^2", 3)
        with self.assertRaises(WrongCaretNotation):
            coords.check_carets()

    def test_offset_without_yaw_and_pitch(self):
        """Checks if offset() method correctly offsets the values."""
        
        coords = Coordinates(1, 2, 3)
        new_coords = coords.offset(1, 1, 1)
        self.assertEqual(new_coords.x, 2)
        self.assertEqual(new_coords.y, 3)
        self.assertEqual(new_coords.z, 4)
        self.assertIsNone(new_coords.yaw)
        self.assertIsNone(new_coords.pitch)

    def test_offset_with_yaw_and_pitch(self):
        """Checks if offset() method correctly offsets the values."""
        
        coords = Coordinates(1, 2, 3, 90, 45)
        new_coords = coords.offset(1, 1, 1, 10, 5)
        self.assertEqual(new_coords.x, 2)
        self.assertEqual(new_coords.y, 3)
        self.assertEqual(new_coords.z, 4)
        self.assertEqual(new_coords.yaw, 100)
        self.assertEqual(new_coords.pitch, 50)

    def test_str_without_yaw_and_pitch(self):
        """Verifies that __str__() method returns the correct string."""
        
        coords = Coordinates(1, 2, 3)
        self.assertEqual(str(coords), "1 2 3")

    def test_str_with_yaw_and_pitch(self):
        """Verifies that __str__() method returns the correct string."""
        
        coords = Coordinates(1, 2, 3, 90, 45)
        self.assertEqual(str(coords), "1 2 3 90 45")

    def test_str_with_caret_notation(self):
        """Verifies that __str__() method returns the correct string."""
        
        coords = Coordinates("^1", "^2", "^3")
        self.assertEqual(str(coords), "^1 ^2 ^3")

    def test_str_with_caret_notation_and_yaw_and_pitch(self):
        """Verifies that __str__() method returns the correct string."""
        
        coords = Coordinates("^1", "^2", "^3", 90, 45)
        self.assertEqual(str(coords), "^1 ^2 ^3 90 45")

    def test_init_with_floats(self):
        """Checks if Coordinates class initializes correctly with floats."""
        
        coords = Coordinates(1.5, 2.5, 3.5)
        self.assertEqual(coords.x, 1.5)
        self.assertEqual(coords.y, 2.5)
        self.assertEqual(coords.z, 3.5)

    def test_init_with_strings(self):
        """Checks if Coordinates class initializes correctly with strings."""
        
        coords = Coordinates("1", "2", "3")
        self.assertEqual(coords.x, "1")
        self.assertEqual(coords.y, "2")
        self.assertEqual(coords.z, "3")

    def test_init_with_mixed_types(self):
        """Checks if Coordinates class initializes correctly with mixed types."""
        
        coords = Coordinates("1", 2.5, 3)
        self.assertEqual(coords.x, "1")
        self.assertEqual(coords.y, 2.5)
        self.assertEqual(coords.z, 3)

    def test_offset_with_floats(self):
        """Checks if offset() method correctly offsets values with floats."""
        
        coords = Coordinates(1.5, 2.5, 3.5)
        new_coords = coords.offset(0.5, 0.5, 0.5)
        self.assertEqual(new_coords.x, 2)
        self.assertEqual(new_coords.y, 3)
        self.assertEqual(new_coords.z, 4)

    def test_str_with_floats(self):
        """Verifies that __str__() method returns the correct string with floats."""
        
        coords = Coordinates(1.5, 2.5, 3.5)
        self.assertEqual(str(coords), "1.5 2.5 3.5")

    def test_str_with_strings(self):
        """Verifies that __str__() method returns the correct string with strings."""
        
        coords = Coordinates("1", "2", "3")
        self.assertEqual(str(coords), "1 2 3")

    def test_str_with_mixed_types(self):
        """Verifies that __str__() method returns the correct string with mixed types."""
        
        coords = Coordinates("1", 2.5, 3)
        self.assertEqual(str(coords), "1 2.5 3")
    
    def test_init_with_negative_values(self):
        """Checks if Coordinates class initializes correctly with negative values."""
        
        coords = Coordinates(-1, -2, -3, -90, -45)
        self.assertEqual(coords.x, -1)
        self.assertEqual(coords.y, -2)
        self.assertEqual(coords.z, -3)
        self.assertEqual(coords.yaw, -90)
        self.assertEqual(coords.pitch, -45)

    def test_offset_with_negative_values(self):
        """Checks if offset() method correctly offsets values with negative values."""
        
        coords = Coordinates(1, 2, 3)
        new_coords = coords.offset(-1, -1, -1)
        self.assertEqual(new_coords.x, 0)
        self.assertEqual(new_coords.y, 1)
        self.assertEqual(new_coords.z, 2)

    def test_init_with_float_yaw_and_pitch(self):
        """Checks if Coordinates class initializes correctly with float yaw and pitch."""
        
        coords = Coordinates(1, 2, 3, 90.5, 45.5)
        self.assertEqual(coords.x, 1)
        self.assertEqual(coords.y, 2)
        self.assertEqual(coords.z, 3)
        self.assertEqual(coords.yaw, 90.5)
        self.assertEqual(coords.pitch, 45.5)

    def test_init_with_string_yaw_and_pitch(self):
        """Checks if Coordinates class initializes correctly with string yaw and pitch."""
        
        coords = Coordinates(1, 2, 3, "90", "45")
        self.assertEqual(coords.x, 1)
        self.assertEqual(coords.y, 2)
        self.assertEqual(coords.z, 3)
        self.assertEqual(coords.yaw, "90")
        self.assertEqual(coords.pitch, "45")




import unittest

from shulker.components.Block import Block
from shulker.components.BlockState import BlockState
from shulker.components.NBT import NBT


class TestBlock(unittest.TestCase):
    def test_block_namespace(self):
        """Test constructor with a valid non-minecraft namespace"""
        block = Block("mod:op_block")

        diff = str(block)
        test = "mod:op_block[]{}"
        self.assertEqual(diff, test)

    def test_block_double_blockstate(self):
        """Test constructor with a valid blockstate of 2 elements"""
        block = Block("bedrock")
        block.blockstate = BlockState({"facing": "north", "half": "top"})

        diff = str(block)
        test = "minecraft:bedrock[facing=north,half=top]{}"
        self.assertEqual(diff, test)

    def test_block_mixed(self):
        """Test constructor with a valid blockstate, and a valid NBT"""
        block = Block("bedrock")
        block.nbt = NBT({"Fire": 4, "Air": 0})
        block.blockstate = BlockState({"facing": "north", "half": "top"})

        diff = str(block)
        test = "minecraft:bedrock[facing=north,half=top]{Fire:4,Air:0}"
        self.assertEqual(diff, test)
            
    def test_constructor_valid_block_id(self):
        """Test constructor with a valid block ID."""
        block = Block("minecraft:stone")
        self.assertEqual(block.descriptor, "minecraft:stone")

    def test_constructor_invalid_block_id(self):
        """Test constructor with an invalid block ID."""
        with self.assertRaises(ValueError):
            Block("minecraft:non_existent_block")
            
        with self.assertRaises(ValueError):
            Block("")

    def test_constructor_default_namespace(self):
        """Test constructor with default namespace (no namespace provided)."""
        block = Block("stone")
        self.assertEqual(block.descriptor, "minecraft:stone")

    def test_constructor_blockstate(self):
        """Test constructor with a valid block state."""
        blockstate = BlockState({"facing": "north"})
        block = Block("minecraft:furnace", blockstate=blockstate)
        self.assertEqual(block.blockstate, blockstate)

    def test_constructor_nbt(self):
        """Test constructor with a valid NBT."""
        nbt = NBT({"BurnTime": 200})
        block = Block("minecraft:furnace", nbt=nbt)
        self.assertEqual(block.nbt, nbt)

    def test_set_blockstate(self):
        """Test setting block state."""
        block = Block("minecraft:furnace")
        block.facing = "north"
        self.assertEqual(block.facing, "north")

    def test_get_blockstate(self):
        """Test getting block state."""
        block = Block("minecraft:furnace")
        block.facing = "north"
        self.assertEqual(block.facing, "north")

    def test_invalid_blockstate(self):
        """Test setting an invalid block state."""
        block = Block("minecraft:furnace")
        with self.assertRaises(ValueError):
            block.invalid_state = "value"

    def test_str_representation(self):
        """Test string representation of a block."""
        with self.assertRaises(ValueError):
            Block("minecraft:furnace[facing=north]{BurnTime:200}")

if __name__ == "__main__":
    unittest.main()



from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.Block import Block
from shulker.components.BlockZone import BlockZone
from shulker.components.Entity import Entity
from shulker.server.singleton import singleton
import sys

__all__ = [
    "check_output_channel",
    "post",
    "format_arg",
    "nest",
]

def check_output_channel():
    if singleton.check_status():
        return singleton
    else:
        raise NoOutputChannelProvided(f"No output channel was initialized in the code")


def post(cmd: str):
    return singleton.post(cmd)


def nesting_process(commands: list, base_coords: BlockCoordinates, set_plane: bool):

    from shulker.functions.default import meta_set_block, meta_summon

    # The activator rail, first passenger, that will get powered
    # once its host the redstone block, has settled on the ground
    rail = Entity('falling_block')
    rail.nbt.id = rail.descriptor
    rail.nbt.BlockState = {'Name':'activator_rail'}
    rail.nbt.Time = 1

    passengers = []

    # This is the pack of third level passengers, this bit of the third level
    # are the custom commands of the nesting
    for command in commands:
      main_cmds_minecart = Entity('command_block_minecart')
      main_cmds_minecart.nbt.id = main_cmds_minecart.descriptor
      main_cmds_minecart.nbt.Command = command
      passengers.append(main_cmds_minecart.nbt)

    # This pack of the third level passengers represents the last
    # part of the nesting, the commands that will kill the contraption

    # This part is the command block that deletes the blocks
    zone = BlockZone(base_coords, base_coords.offset(0, -2 if not set_plane else -3, 0))
    block = Block('command_block')
    block.nbt = {'auto':1, 'Command':f'fill {zone} air'}

    command = meta_set_block(base_coords, block, "replace")

    deleter_minecart = Entity('command_block_minecart')
    deleter_minecart.nbt.id = deleter_minecart.descriptor
    deleter_minecart.nbt.Command = command
    passengers.append(deleter_minecart.nbt)

    # This part is the command block that kills the contraption entities
    command = 'kill @e[distance=..1]'

    killer_minecart = Entity('command_block_minecart')
    killer_minecart.nbt.id = killer_minecart.descriptor
    killer_minecart.nbt.Command = command
    passengers.append(killer_minecart.nbt)

    rail.nbt.Passengers = passengers

    # The base of the contraption, the redstone block that will power
    # its next passenger, the activator rail.
    base = Entity("falling_block")
    base.nbt.BlockState = {'Name':'redstone_block'}
    base.nbt.Time = 1
    base.nbt.Passengers = [rail.nbt]

    cmd = meta_summon(base.descriptor, base_coords.offset(0, 0, 0), base.nbt)
    
    return cmd


def nest_commands(commands: list, base_coords: BlockCoordinates, set_plane):
  
    # NOTE: The byte size of the string also depends on using do
    # double or single quotes 
    
    # This is 49, the size of an empty string
    string_overhead = sys.getsizeof('')
    
    # This is 88, the size of a string with the value of an empty command block minecart
    # We substract the overhead because it will be put inside the command
    string_addition = sys.getsizeof('{id:"command_block_minecart",Command:""},') - string_overhead
    
    # This is 356, the size of the base command if you substract the empty command block minecart
    # it is substracted because it will be added for each commands
    cmd = nesting_process([""], base_coords, set_plane)
    base_cmd_size = sys.getsizeof(cmd) - string_addition
    
    # We save the base size, and create a variable that will be used to
    # track the current size of the nest we're creating (= pkg/package)
    current_cmd_size = base_cmd_size
    
    nests = []
    pkg = []
    
    # While there are still any commands left to be nested
    while len(commands):
      
      # Determine the size of the next command
      next_command_size = sys.getsizeof(commands[0]) + string_addition - string_overhead
      
      # If it fits in the current nest, add it to the nest (pkg)
      if current_cmd_size + next_command_size < 1400:
        current_cmd_size += next_command_size
        pkg.append(commands.pop(0))
      
      # Otherwise, create a new nest (pkg) and add the previous nest in the list of nests
      else:
        current_cmd_size = base_cmd_size
        nests.append(pkg)
        pkg = []
    else:
      # Add the last nest to the list of nests
      nests.append(pkg)
    
    return nests


def place_nests(nests: list, base_coords: BlockCoordinates, set_plane, plane_block):
  
  from shulker.functions.default import meta_set_zone

  # If set_plane is set to true, we fill a zone with a plane that will
  # be used to place the nests on top of it
  
  cmds = []
  
  if set_plane:
    wideness = round(len(nests) / 16)
    zone = BlockZone(base_coords.offset(0, -3, 0), base_coords.offset(0 + wideness, -3, len(nests) - 1 % 16))
    cmd = meta_set_zone(zone, plane_block, "replace")
    cmds.append(cmd)
  
  for i, nest in enumerate(nests):
    cmd = nesting_process(nest, base_coords.offset(round(i/16), 0, i % 16), set_plane)
    cmds.append(cmd)

  return cmds
  
  
def nest(commands: list, base_coords: BlockCoordinates, set_plane=True, plane_block=Block('barrier'), dry_run=False):
  nests = nest_commands(commands, base_coords, set_plane)
  
  cmds = place_nests(nests, base_coords, set_plane, plane_block)
  
  if dry_run:
    return cmds
  else:
    for cmd in cmds:
      post(cmd)


def format_arg(argument, component):
    if isinstance(argument, component):
        return argument

    if isinstance(argument, str):
        return component(argument)
    elif isinstance(argument, tuple):
        return component(*argument)
    elif isinstance(argument, list):
        return component(*argument)
    else:
        return argument
        raise InvalidArgumentType(
            f'Invalid argument type, could not parse "{argument}" as a valid component.'
        )


class InvalidArgumentType(Exception):
    pass


class UnexpectedReturn(Exception):
    pass


class NoOutputChannelProvided(Exception):
    pass



import unittest

from shulker.components.Coordinates import WrongCaretNotation
from shulker.components.BlockCoordinates import BlockCoordinates


class TestCoordinates(unittest.TestCase):
    def test_blockcoords_zeros(self):
        """Test that initializing BlockCoordinates with all zeros produces the expected output."""
        coords = BlockCoordinates(0, 0, 0)

        diff = str(coords)
        test = "0 0 0"
        self.assertEqual(diff, test)

    def test_coords_tilde(self):
        """Test that initializing BlockCoordinates with all tildes produces the expected output."""
        coords = BlockCoordinates("~", "~", "~")

        diff = str(coords)
        test = "~ ~ ~"
        self.assertEqual(diff, test)

    def test_coords_tilde_mixed(self):
        """Test that initializing BlockCoordinates with a mix of tildes and numbers produces the expected output."""
        coords = BlockCoordinates("~5", 0, "~-10")

        diff = str(coords)
        test = "~5 0 ~-10"
        self.assertEqual(diff, test)

    def test_coords_caret(self):
        """Test that initializing BlockCoordinates with all carets produces the expected output."""
        coords = BlockCoordinates("^", "^", "^")

        diff = str(coords)
        test = "^ ^ ^"
        self.assertEqual(diff, test)

    def test_coords_caret_mixed(self):
        """Test that initializing BlockCoordinates with a mix of carets and numbers produces the expected output."""
        coords = BlockCoordinates("^5", "^-10", "^")

        diff = str(coords)
        test = "^5 ^-10 ^"
        self.assertEqual(diff, test)

    def test_coords_caret_wrong(self):
        """Test that initializing BlockCoordinates with an incorrect caret notation raises the expected exception."""
        coords = BlockCoordinates(5, "^", "^")

        with self.assertRaises(WrongCaretNotation):
            str(coords)

    def test_coords_floats(self):
        """Test that initializing BlockCoordinates with floats and mixed notations produces the expected output."""
        coords = BlockCoordinates(1.5, "~0.1", 1)

        diff = str(coords)
        test = "2 ~0 1"
        self.assertEqual(diff, test)

    def test_blockcoords_offset(self):
        """Test that the offset method of BlockCoordinates produces the expected output."""
        coords = BlockCoordinates(1, 2, 3)
        offset_coords = coords.offset(1, 1, 1)
        
        diff = str(offset_coords)
        test = "2 3 4"
        self.assertEqual(diff, test)
    
if __name__ == "__main__":
    unittest.main()



import math
import re

from .Coordinates import Coordinates

class BlockCoordinates(Coordinates):
    """
    The position of a block is actually the coordinates of the point at the lower northwest corner
    of the block, that is, the integer coordinates obtained by rounding down the coordinates
    inside the block.

    In Minecraft, decimal coordinates usually needs to be converted into integer coordinates by rounding
    down, which is called the block position of the coordinate.
    """

    def round_coord(self, coord):
        if isinstance(coord, str) and "." in coord:
            coord = re.sub(r"([\^~])(\d+\.\d+)", lambda m: m.group(1) + str(round(float(m.group(2)))), coord)
        elif isinstance(coord, float):
            coord = round(coord)
        return coord

    def offset(self, x=0, y=0, z=0) -> "BlockCoordinates":
        """
        Offsets the coordinates by the given tuple and returns a new BlockCoordinates object.
        """
        
        return BlockCoordinates(self.x + x, self.y + y, self.z + z)
    
    def set(self, block, handler=None):
        from shulker.functions.default import set_block
        from .Block import Block
        
        self.__dict__['block'] = Block(block)
        
        if handler:
            return set_block(self, self.__dict__['block'], handler)
        else:
            return set_block(self, self.__dict__['block'])
        
    def __str__(self):

        self.check_carets()

        self.x = self.round_coord(self.x)
        self.y = self.round_coord(self.y)
        self.z = self.round_coord(self.z)

        return f"{self.x} {self.y} {self.z}"



import unittest

from shulker.components.Coordinates import Coordinates
from shulker.components.Zone import Zone, ZoneWrongCoordsType


class TestCoordinates(unittest.TestCase):
    def test_zone_with_coordinates(self):
        """Test if Zone works with Coordinates instances for both positions."""
        coords1 = Coordinates(0, 0, 0)
        coords2 = Coordinates(10, 10, 10)
        zone = Zone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_with_tuple(self):
        """Test if Zone works with tuples for both positions."""
        coords1 = (0, 0, 0)
        coords2 = (10, 10, 10)
        zone = Zone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_mixed(self):
        """Test if Zone works with a mix of Coordinates instance and tuple for positions."""
        coords1 = (0, 0, 0)
        coords2 = Coordinates(10, 10, 10)
        zone = Zone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_wrong_args(self):
        """Test if Zone raises ZoneWrongCoordsType when provided with invalid arguments."""
        zone = Zone("hello", 1)
        with self.assertRaises(ZoneWrongCoordsType):
            str(zone)

    def test_zone_contains_true(self):
        """Test if Zone contains method returns True for a coordinate within the zone."""
        coords1 = Coordinates(0, 0, 0)
        coords2 = Coordinates(10, 10, 10)
        zone = Zone(coords1, coords2)
        test_coord = Coordinates(5, 5, 5)

        self.assertTrue(zone.contains(test_coord))

    def test_zone_contains_false(self):
        """Test if Zone contains method returns False for a coordinate outside the zone."""
        coords1 = Coordinates(0, 0, 0)
        coords2 = Coordinates(10, 10, 10)
        zone = Zone(coords1, coords2)
        test_coord = Coordinates(15, 15, 15)

        self.assertFalse(zone.contains(test_coord))

    def test_zone_contains_error(self):
        """Test if Zone contains method raises ZoneWrongCoordsType for non-Coordinates input."""
        coords1 = Coordinates(0, 0, 0)
        coords2 = Coordinates(10, 10, 10)
        zone = Zone(coords1, coords2)

        with self.assertRaises(ZoneWrongCoordsType):
            zone.contains("hello")

    
if __name__ == "__main__":
    unittest.main()



import unittest

from nbtlib import serialize_tag, Compound

from shulker.components.NBT import NBT
from shulker.components.Block import Block

class TestBlock(unittest.TestCase):
    def test_init_empty(self):
        """Test empty NBT object initialization."""
        nbt = NBT()
        self.assertEqual(str(nbt), "{}")

    def test_init_compound(self):
        """Test NBT object initialization with Compound input."""
        compound = Compound({"test_key": "test_value"})
        nbt = NBT(compound)
        self.assertEqual(str(nbt), '{test_key:"test_value"}')

    def test_init_str(self):
        """Test NBT object initialization with string input."""
        nbt_str = '{"test_key": "test_value"}'
        nbt = NBT(nbt_str)
        self.assertEqual(str(nbt), '{test_key:"test_value"}')

    def test_init_dict(self):
        """Test NBT object initialization with dictionary input."""
        nbt_dict = {"test_key": "test_value"}
        nbt = NBT(nbt_dict)
        self.assertEqual(str(nbt), '{test_key:"test_value"}')

    def test_init_invalid_type(self):
        """Test ValueError is raised with invalid input type."""
        with self.assertRaises(ValueError):
            NBT(["invalid", "input"])

    def test_setattr(self):
        """Test attribute setting and conversion to NBT."""
        nbt = NBT()
        nbt.test_key = "test_value"
        self.assertEqual(str(nbt), '{test_key:"test_value"}')

    def test_str(self):
        """Test NBT object serialization to string."""
        nbt = NBT()
        nbt.test_key = "test_value"
        nbt_repr = str(nbt)
        self.assertEqual(nbt_repr, '{test_key:"test_value"}')

    def test_set_attribute_different_types(self):
        """Test setting attributes of different types on an NBT object."""
        nbt = NBT()
        nbt.int_value = 42
        nbt.float_value = 3.14
        nbt.str_value = "hello"
        self.assertEqual(serialize_tag(nbt.int_value, compact=True), "42")
        self.assertEqual(serialize_tag(nbt.float_value, compact=True), "3.14d")
        self.assertEqual(serialize_tag(nbt.str_value, compact=True), '"hello"')

    def test_update_attribute(self):
        """Test updating an existing attribute on an NBT object."""
        nbt = NBT({"key": "value"})
        nbt.key = "new_value"
        self.assertEqual(str(nbt), '{key:"new_value"}')

    def test_delete_attribute(self):
        """Test deleting an attribute from an NBT object."""
        nbt = NBT({"key": "value", "key2": "value2"})
        del nbt.key
        self.assertEqual(str(nbt), '{key2:"value2"}')

    def test_delete_attribute_error(self):
        """Test AttributeError when deleting a non-existent attribute."""
        nbt = NBT()
        with self.assertRaises(AttributeError):
            del nbt.non_existent_key

    def test_nested_nbt_init(self):
        """Test initializing an NBT object with nested NBT data."""
        nested_nbt_data = {"key": {"nested_key": "nested_value"}}
        nbt = NBT(nested_nbt_data)
        self.assertEqual(str(nbt), '{key:{nested_key:"nested_value"}}')

    def test_nested_nbt_set_attribute(self):
        """Test setting a nested attribute on an NBT object."""
        
        nbt = NBT()
        nbt.key = NBT({"nested_key": "nested_value"})
        self.assertEqual(str(nbt), '{key:{nested_key:"nested_value"}}')

    def test_nested_nbt_update_attribute(self):
        """Test updating a nested attribute on an NBT object."""
        
        nbt = NBT({"key": NBT({"nested_key": "nested_value"})})
        nbt.key.nested_key = "new_nested_value"
        self.assertEqual(str(nbt), '{key:{nested_key:"new_nested_value"}}')
    
if __name__ == "__main__":
    unittest.main()


import unittest

from nbtlib import serialize_tag

from shulker.components.Entity import Entity, NBT, entities_nbt

class TestEntity(unittest.TestCase):

    def test_invalid_entity(self):
        """Test Entity class raises ValueError with invalid entity name."""
        with self.assertRaises(ValueError):
            Entity("invalid_entity")

    def test_valid_entity(self):
        """Test Entity class creates a valid instance with a valid entity name."""
        entity = Entity("Zombie")
        self.assertEqual(entity.descriptor, "zombie")

    def test_nbt_type(self):
        """Test Entity class raises ValueError with invalid nbt input type."""
        with self.assertRaises(ValueError):
            Entity("zombie", nbt="invalid_nbt_type")

    def test_valid_nbt(self):
        """Test Entity class creates a valid instance with a valid NBT object input."""
        entity = Entity("zombie", nbt=NBT({"CustomName": "zombie1"}))
        self.assertEqual(serialize_tag(entity.CustomName, compact=True), '"zombie1"')

    def test_dict_nbt(self):
        """Test Entity class creates a valid instance with a valid dictionary input for nbt."""
        entity = Entity("zombie", nbt={"CustomName": "zombie2"})
        self.assertEqual(serialize_tag(entity.CustomName, compact=True), '"zombie2"')

    def test_set_nonexistent_attribute(self):
        """Test Entity class raises ValueError when trying to set a nonexistent attribute to an instance."""
        entity = Entity("zombie")
        with self.assertRaises(ValueError):
            entity.invalid_attribute = "value"

    def test_set_valid_attribute(self):
        """Test Entity class sets a valid attribute for an instance."""
        entity = Entity("zombie")
        entity.Health = 20
        self.assertEqual(serialize_tag(entity.Health, compact=True), '20')

    def test_uuid_creation(self):
        """Test Entity class generates a UUID for each instance."""
        import re
        entity = Entity("zombie")
        uuid_str = entity.UUID
        uuid_list = [int(num) for num in re.findall(r'-?\d+', serialize_tag(uuid_str, compact=True))]
        self.assertEqual(len(uuid_list), 4)
        for i in uuid_list:
            self.assertTrue(isinstance(i, int))

    def test_str_representation(self):
        """Test __str__ method of the Entity class returns a string representation of the entity."""
        entity = Entity("zombie")
        self.assertTrue(isinstance(entity.__str__(), str))

    def test_attributes_property(self):
        """Test Entity class returns expected attributes for an entity."""
        entity = Entity("zombie")
        expected_attributes = entities_nbt["zombie"]
        self.assertEqual(entity.attributes, expected_attributes)




from typing import Union

from shulker.components.NBT import NBT
from shulker.components.TargetSelector import TargetSelector
import shulker.functions.base_functions

def meta_update(selector: TargetSelector, nbt: NBT) -> str:
    return f"data merge entity {selector} {nbt}"

def update(
    selector: Union[TargetSelector, str],
    nbt: Union[NBT, dict, str]
) -> str:
    """
    Updates an entity with nbt_data
    
    selector can either be a TargetSelector object or an UUID string gotten
    from the UUID nbt of an entity
    """
    
    shulker.functions.base_functions.check_output_channel()
        
    if type(nbt) is dict:
        nbt = NBT(nbt)
    elif type(nbt) is str:
        nbt = NBT(nbt)
    elif type(nbt) is not NBT:
        raise ValueError(f"nbt must be of type NBT, dict, or str, not {type(nbt)}")
    
    if type(selector) is str:
        selector = TargetSelector("e", {"nbt": f"{{UUID:{selector}}}", "limit": 1})
        
    cmd = meta_update(selector, nbt)
    
    status = shulker.functions.base_functions.post(cmd)
    
    return status


import unittest

from shulker.components.BlockHandler import BlockHandler, BlockHandlerWrongType
from shulker.components.Block import Block


class TestBlockHandler(unittest.TestCase):
    def test_handler_destroy(self):
        """Test if the destroy option is properly set and returned."""
        bh = BlockHandler("destroy")

        diff = str(bh)
        test = "destroy"
        self.assertEqual(diff, test)

    def test_handler_replace(self):
        """Test if the replace option is properly set and returned."""
        bh = BlockHandler("replace")

        diff = str(bh)
        test = "replace"
        self.assertEqual(diff, test)

    def test_handler_keep(self):
        """Test if the keep option is properly set and returned."""
        bh = BlockHandler("keep")

        diff = str(bh)
        test = "keep"
        self.assertEqual(diff, test)

    def test_handler_hollow(self):
        """Test if the hollow option is properly set and returned."""
        bh = BlockHandler("hollow")

        diff = str(bh)
        test = "hollow"
        self.assertEqual(diff, test)

    def test_handler_outline(self):
        """Test if the outline option is properly set and returned."""
        bh = BlockHandler("outline")

        diff = str(bh)
        test = "outline"
        self.assertEqual(diff, test)

    def test_handler_default(self):
        """Test if the default option (replace) is properly set and returned."""
        bh = BlockHandler()

        diff = str(bh)
        test = "replace"
        self.assertEqual(diff, test)

    def test_abritrary_handler(self):
        """Test if an arbitrary option raises the BlockHandlerWrongType exception."""
        with self.assertRaises(BlockHandlerWrongType):
            BlockHandler("imaginary")

    def test_wrong_handler_type(self):
        """Test if an incorrect option type (non-string) raises the BlockHandlerWrongType exception."""
        with self.assertRaises(BlockHandlerWrongType):
            BlockHandler(1)


if __name__ == "__main__":
    unittest.main()



from typing import Union

from .BlockCoordinates import BlockCoordinates

class BlockZone:
    """
    Custom component that is a set of two BlockCoordinates(), representing an area
    """

    def __init__(
        self, pos1: Union[BlockCoordinates, tuple], pos2: Union[BlockCoordinates, tuple]
    ):

        if isinstance(pos1, tuple):
            self.pos1 = BlockCoordinates(*pos1)
        else:
            self.pos1 = pos1

        if isinstance(pos2, tuple):
            self.pos2 = BlockCoordinates(*pos2)
        else:
            self.pos2 = pos2

    def contains(self, coord: BlockCoordinates) -> bool:
        """
        Checks if the given BlockCoordinates is within the BlockZone.
        """
        if not isinstance(coord, BlockCoordinates):
            raise BlockZoneWrongCoordsType(
                "BlockZone contains method requires a BlockCoordinates instance as input"
            )

        min_x = min(self.pos1.x, self.pos2.x)
        max_x = max(self.pos1.x, self.pos2.x)
        min_y = min(self.pos1.y, self.pos2.y)
        max_y = max(self.pos1.y, self.pos2.y)
        min_z = min(self.pos1.z, self.pos2.z)
        max_z = max(self.pos1.z, self.pos2.z)

        return (min_x <= coord.x <= max_x) and (min_y <= coord.y <= max_y) and (min_z <= coord.z <= max_z)

    def set(self, block, handler="replace", filter=""):
        from shulker.functions.default import set_zone

        self.__dict__['block'] = block
        return set_zone(self, block, handler, filter)
    
    def move(self, zone, handler="replace", filter="", replace_by="air"):
        from shulker.functions.default import set_zone

        if 'block' not in self.__dict__:
            raise ValueError("BlockZone.move() requires a block to be set first (by .set())")
        
        old_zone = BlockZone(self.pos1, self.pos2)
        
        if type(zone) is not BlockZone:
            new_zone = BlockZone(zone)
        else:
            new_zone = zone
        
        self.pos1 = new_zone.pos1
        self.pos2 = new_zone.pos2
        
        set_zone(old_zone, replace_by, handler, filter)
        set_zone(self, self.__dict__['block'], handler, filter)
    
    def offset_place(self, x=0, y=0, z=0, handler='replace', filter="", replace_by="air"):
        new_zone = self.offset(x, y, z)
        self.move(new_zone, handler=handler, filter=filter, replace_by=replace_by)
    
    def offset(self, x=0, y=0, z=0):
        """
        Offsets the coordinates by the given tuple and returns a new Zone object.
        """
        return BlockZone(self.pos1.offset(x, y, z), self.pos2.offset(x, y, z))
        
    def __str__(self):
        if not isinstance(self.pos1, BlockCoordinates) or not isinstance(
            self.pos2, BlockCoordinates
        ):
            raise BlockZoneWrongCoordsType(
                "a BlockZone requires to be provided a set of two BlockCoordinates instances as pos1 and pos2"
            )

        return f"{self.pos1} {self.pos2}"


class BlockZoneWrongCoordsType(Exception):
    pass



import unittest

from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.BlockZone import BlockZone, BlockZoneWrongCoordsType


class TestCoordinates(unittest.TestCase):
    def test_zone_with_blockcoords(self):
        """Test if BlockZone works with BlockCoordinates instances for both positions."""
        coords1 = BlockCoordinates(0, 0, 0)
        coords2 = BlockCoordinates(10, 10, 10)
        zone = BlockZone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_with_tuple(self):
        """Test if BlockZone works with tuples for both positions."""
        coords1 = (0, 0, 0)
        coords2 = (10, 10, 10)
        zone = BlockZone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_mixed(self):
        """Test if BlockZone works with a mix of BlockCoordinates instance and tuple for positions."""
        coords1 = (0, 0, 0)
        coords2 = BlockCoordinates(10, 10, 10)
        zone = BlockZone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_wrong_args(self):
        """Test if BlockZone raises ZoneWrongCoordsType when provided with invalid arguments."""
        zone = BlockZone("hello", 1)
        with self.assertRaises(BlockZoneWrongCoordsType):
            str(zone)

    def test_zone_with_negative_coords(self):
        """Test if BlockZone works with negative coordinates."""
        coords1 = BlockCoordinates(-5, 0, -5)
        coords2 = BlockCoordinates(5, 5, 5)
        zone = BlockZone(coords1, coords2)

        diff = str(zone)
        test = "-5 0 -5 5 5 5"
        self.assertEqual(diff, test)
        
    def test_zone_same_coordinates(self):
        """Test if BlockZone works with the same coordinates for both positions."""
        coords1 = BlockCoordinates(0, 0, 0)
        coords2 = BlockCoordinates(0, 0, 0)
        zone = BlockZone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 0 0 0"
        self.assertEqual(diff, test)

    def test_zone_larger_coords_in_pos1(self):
        """Test if BlockZone works with larger coordinates in position 1 than in position 2."""
        coords1 = BlockCoordinates(10, 10, 10)
        coords2 = BlockCoordinates(0, 0, 0)
        zone = BlockZone(coords1, coords2)

        diff = str(zone)
        test = "10 10 10 0 0 0"
        self.assertEqual(diff, test)

    def test_zone_float_coordinates_rounded(self):
        """Test if BlockZone rounds floating-point coordinates correctly."""
        coords1 = (0.5, 1.2, 0)
        coords2 = (10.8, 10.3, 10.1)
        zone = BlockZone(coords1, coords2)

        diff = str(zone)
        test = "0 1 0 11 10 10"
        self.assertEqual(diff, test)


    
if __name__ == "__main__":
    unittest.main()



import json
from typing import Union

from nbtlib import parse_nbt, serialize_tag
from nbtlib import Compound, List, Int

class NBT:
    def __init__(self, compound: Union[dict, str, Compound] = "{}"):

        if compound in [None, "", "{}", {}]:
            nbt = None
        elif isinstance(compound, Compound):
            nbt = compound
        elif isinstance(compound, str):
            nbt = json.loads(compound)
        elif isinstance(compound, dict):
            nbt = Compound(compound)
        else:
            raise ValueError(f"Expected dict, str or Compound, got {type(compound)}")

        if nbt:
            for key, value in nbt.items():
                if isinstance(value, Compound):
                    setattr(self, key, NBT(value))
                else:
                    setattr(self, key, value)

    def __setattr__(self, name, value):
        if isinstance(value, NBT):
            super().__setattr__(name, value)
        else:
            nbt = parse_nbt(str({name: value}))
            for key in nbt:
                super().__setattr__(key, nbt[key])

    def __str__(self):
        """The 'list of int' hacky thing is to fix a
        missing 'I;' in the serialized NBT tag.
        That minecraft seems to require"""
        
        compound = Compound()
        
        is_list_of_int = []
        for key, value in vars(self).items():
            if isinstance(value, NBT):
                compound[key] = Compound(value.__dict__)
            else:
                if isinstance(value, list):
                    if all([isinstance(x, int) for x in value]):
                        is_list_of_int.append(key)
                compound[key] = value
        
        serialized = serialize_tag(compound, compact=True)
        
        for list_of_int in is_list_of_int:
            serialized = serialized.replace(f"{list_of_int}:[", f"{list_of_int}:[I;")
        
        return serialized